#!/usr/bin/env python

##############################################################################
##
## Copyright (C) 2021 Narda Safety Test Solutions GmbH.
## Contact: www.narda-sts.com
##
## Redistribution and use in source and binary forms,
## with or without modification, are permitted provided that the following
## conditions are met:
##
## 1.) Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2.) Redistributions in binary form must reproduce the above copyright
## notice, this list of conditions and the following disclaimer in the
## documentation and/or other materials provided with the distribution.
##
## 3.) Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from this
## software without specific prior written permission.
##
## 4.) EXPORT CONTROL
## The Software, including technical data /cryptographic software,
## may be subject to, German, European Union and U.S. export controls and
## may be subject to import or export controls in other countries.
## The Licensee agrees to strictly comply with all applicable import and
## export regulations. He specifically agrees, that he must not disclose or
## otherwise export or re-export the Licensed Software or any part thereof
## delivered under this end user license agreement (EULA) to any country
## (including a national or resident of such country) without a valid export
## or import license. Please be aware that the Software may contain
## US-Content, therefor the Licensee represent and warrant that he is not a
## citizen of, or otherwise located within, an embargoed nation (including
## without limitation Iran, Syria, Sudan, Cuba, North Korea) and that he is
## not otherwise prohibited under the Export Laws from receiving the Software.
## All rights to Use the Software are granted on condition that such rights
## are forfeited if the Licensee fails to comply with
## the terms of this Agreement.
##
## 5.) SEVERABILITY
## Should any provision of this Agreement be or become invalid, ineffective
## or unenforceable, the remaining provisions of this Agreement shall be
## valid. The parties agree to replace the invalid, ineffective or
## unenforceable provision by a valid, effective and enforceable provision
## which best meets the commercial intention of the parties.
## The same shall apply in case of omissions.
##
## 6.) APPLICABLE LAW AND PLACE OF JURISTICATION
## This Agreement shall be constituted under the law of Germany.
## The United Nations Convention on the International Sale of Goods
## shall not apply to this Agreement. The Place of Jurisdiction for any
## dispute between the Parties shall be Tübingen (Germany).
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
## TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Narda Safety Test Solutions GmbH
## BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##
##############################################################################


from pathlib import Path
import wx
from nardascripting.base.usrscriptbase import *
from nardascripting.base.signalsharkdev import *
from nardascripting.base.measthread import *
from .cmapsettings import CMapSettings
from .gui.cmapdlgs import LevelData, LevelScale, LevelScaleItem, MeasParameters
import time
import csv
from lxml import etree
from pykml.parser import Schema
from pykml.factory import KML_ElementMaker as KML
from pykml.factory import GX_ElementMaker as GX


APP_TITLE ="Convert CMap CSV to KML V1.0"


class ConvertCMapToKML(UsrScriptBase):
    """User script class"""

    def __init__(self, main_gui, dev=SignalSharkDev()):
        """Initialization. Please leave this line unchanged"""
        super().__init__(main_gui, dev, __file__)

        # Script settings.
        # -------------------------------------------------------------------------------
        # Please adapt the following lines of code according to your script.
        self._tab_name = "nslcoveragemap"
        self._scr_title = APP_TITLE
        self._scr_description = "Converts a CMap CSV file to KML."
        # self._icon_path = self.script_path.joinpath("NardaScriptLauncher_example01.png")
        self._list_prio = 2
        self._nsl_executed_behavior = NSL_Executed_Behaviors.MINIMIZE_NSL

        # Add class variables if needed
        # -------------------------------------------------------------------------------
        self.settings = {}

    def _run_script(self, args):
        """
        Script main function
        This method is called when a user clicks on the corresponding script button.
        """

        # Load coverage map settings
        # -------------------------------------------------------------------------------
        self.settings = CMapSettings.load()
        if not self.settings:
            self.MessageBoxModal("Please run App Coverage Map Settings first!",
                                 "Coverage Map", wx.OK | wx.ICON_INFORMATION)
            return

        # Get Coverage Map CSV file from Open File Dialog:
        # -------------------------------------------------------------------------------
        dlg = wx.FileDialog(self.frm_dlghelper, "Open CMap CSV file", "", "",
                            "CMap CSV files (*.csv)|*.csv",
                            wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)

        if self.ShowDlgModalTop(dlg) != wx.ID_OK:
            dlg.Destroy()
            return
        file_name = dlg.GetPath()
        dlg.Destroy()  # Please do not forget to destroy the created dialog if you do not need it anymore


        # Create an additional measurement thread to keep the GUI alive.
        # With "args=[param1, parm2, paramn]" you can pass some optional parameters to the measurement thread.
        mthread = MeasDlgThread(main_gui=self.main_gui, callback=ConvertCMapToKML._run_measurement,
                                args=[self.settings, file_name])

        # Start the measurement thread and show a visualization (progress) dialog.
        dlg_result = mthread.start_measurement()
        if dlg_result == wx.ID_OK:
            self.main_gui.MessageBoxModal("Conversion done.", "Info")
        else:
            self.main_gui.MessageBoxModal("Conversion canceled", "Info")

    @staticmethod
    def _run_measurement(stopevent, update_progress, wait_for_stopevent,
                         cmap_settings: dict, file_name: str):
        """ Main method of the measurement thread that handles a time consuming measurement.

        :param stopevent: Thread flag that indicates whether the procedure should be finished/cancelled
        :param update_progress: Delegate method to update a progress bar, message string and icon
        :param wait_for_stopevent: Delegate method to wait until stopevent is raised with possibility to show a message.
        :param cmap_settings: Custom parameter to handover the coverage map settings
        :param file_name: Custom parameter to handover the file name as string.
        :return:
        """
        # Initialize return value (if any)
        succeeded = False

        # Prepare progress value calculation. Progress is a value from 0.0 to 1.0
        steps = 0.0
        data_file = None
        try:
            # Update the dialog message text and show a progress bar
            # -------------------------------------------------------------------------------
            update_progress(msg="Loading data....",
                            show_progress=True, btn_style=wx.OK)

            settings_unit = cmap_settings.get("unit", "")

            # Create level scale:
            # -------------------------------------------------------------------------------
            level_scale_dict = cmap_settings.get("level_scale", None)
            if level_scale_dict:
                level_scale = LevelScale.from_dict(level_scale_dict)
            else:
                level_scale = LevelScale.get_default()

            # Get measurement parameters:
            # -------------------------------------------------------------------------------
            meas_param = MeasParameters.from_csv_file(file_name)
            description = "Input.: {}<br>" \
                          "Atten.: {} dB<br>" \
                          "Unit: {}<br>" \
                          "Ftune: {} MHz<br>" \
                          "CBW: {} MHz<br>" \
                          "Det. Type: {}<br>" \
                          "Meas. Time: {} s<br>" \
                          "Post Avg.: {} s<br>" \
                          "Pre Avg.: {} s<br>" \
                          "Comment: {}".format(meas_param.input, meas_param.atten, meas_param.unit,
                                               meas_param.ftune/1e6, meas_param.cbw/1e6, meas_param.det,
                                               meas_param.meas_time, meas_param.post_avg, meas_param.pre_avg,
                                               meas_param.comment)

            if meas_param.unit != settings_unit:
                return wait_for_stopevent(
                    "The unit of the measurement file does not match the unit of the settings file!")


            # Open csv file:
            # -------------------------------------------------------------------------------
            data_file = open(file_name, "r")
            data_reader = csv.reader(data_file, delimiter=";")

            # Get number of lines for progress bar
            lines = sum(1 for _ in data_reader)
            data_file.seek(0)
            progress = 0.0
            progress_step = 1.0 / lines

            # Show Message "Conversion in progress"
            update_progress(msg="Conversion in progress", progress=0.0)


            # Create new KML file:
            # -------------------------------------------------------------------------------
            csv_file_path = Path(file_name)
            meas_name = csv_file_path.name.replace(".csv", "")
            kml_file_name = csv_file_path.parent.joinpath(meas_name + ".kml")
            kmlobj = KML.kml(KML.Document(KML.name("Narda CMap Data")))

            # Add level scale as styles:
            # -------------------------------------------------------------------------------
            for item in level_scale:
                kml_colour = ConvertCMapToKML._get_kml_color(item.colour)
                ConvertCMapToKML._add_style(kmlobj, kml_colour)
            last_item = level_scale.get_last_item()
            if last_item:
                kml_colour = ConvertCMapToKML._get_kml_color(level_scale.base_color)
                ConvertCMapToKML._add_style(kmlobj, kml_colour)

            # Add folder for placemarks:
            # -------------------------------------------------------------------------------
            kml_folder = KML.Folder(KML.name(meas_name + "-" + meas_param.det), KML.description(description))

            # Skip header
            for i in range(18):
                data_file.readline()

            check_cnt = 0
            line = 18
            for row in data_reader:
                # Periodically check, whether the user has clicked the cancel button,
                # and exit thread if so.
                if check_cnt >= 50:
                    # Allow the thread to synchronize the stop event.
                    stopevent.wait(0.1)
                    if stopevent.is_set():
                        return
                    check_cnt = 0
                    update_progress(progress=progress)

                data = LevelData()
                data.timestamp = row[0]
                data.gnss = GnssCoordinate(float(row[1]), float(row[2]))
                data.det_val = float(row[4])
                kml_colour = ConvertCMapToKML._get_kml_color(level_scale.get_lev_color(data.det_val))

                # Convert to KML
                # -------------------------------------------------------------------------------
                name = "{}: {} {}".format(meas_param.det, row[4], meas_param.unit)
                style_name = "#stylecmap_" + kml_colour
                ConvertCMapToKML._add_placemark(kml_folder, name,
                                                GnssCoordinate(float(row[1]), float(row[2])),
                                                style_name, "Item{}".format(line))

                progress += progress_step
                check_cnt += 1
                line += 1
            kmlobj.Document.append(kml_folder)

            # Write KML file
            # -------------------------------------------------------------------------------
            with open(kml_file_name, "w") as f:
                f.write(str(etree.tostring(kmlobj, encoding='unicode', pretty_print=True)))

        except Exception as e:
            # Do some error handling
            return wait_for_stopevent(str(e), icon_style=wx.ICON_ERROR)
        finally:
            # Close file and connection.
            if data_file:
                data_file.close()

    @staticmethod
    def _get_kml_color(colour: wx.Colour):
        """Calculates kml colour (AABBGGRR) from wx.Colour"""
        retvalue = "ff{:02x}{:02x}{:02x}".format(colour.Blue(), colour.Green(), colour.Red())
        return retvalue

    @staticmethod
    def _add_style(kmlobj, kml_colour: str):
        """Adds a colour style"""
        icon_id = "iconcmap_" + kml_colour
        style_id = "stylecmap_" + kml_colour
        new_style = KML.Style(
            KML.IconStyle(
                KML.color(kml_colour),
                KML.scale(1.0),
                KML.Icon(
                    KML.href("http://maps.google.com/mapfiles/kml/shapes/road_shield3.png"),
                ),
                id=icon_id
            ),
            id=style_id
        )
        kmlobj.Document.append(new_style)

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _add_placemark(kml_folder, name: str, gnss: GnssCoordinate, style: str, id: str):
        kml_folder.append(KML.Placemark(KML.name(name),
                                 KML.styleUrl(style),
                                 KML.Point(KML.coordinates(gnss.longitude, ",", gnss.latitude)),
                                 id=id))
